// Implementaion file for graph class.

#include "graph.h"

bool isIn(vector<int> fb, int i);



graph::graph()
{
	sourceNodeID = 0;
	targetNodeID = 0;
}

graph::~graph()
{

}

graph::graph(int x, int y, int fCost, int bCost)
{
	sourceNodeID = x;
	targetNodeID = y;
	forwardCost = fCost;
	backwardCost = bCost;
}


void graph::addNode(node nd)
{
	allNodes.push_back(nd);
}

void graph::show ()
{

	cout <<endl ;
	cout << "Following is the adjacency list representation of the graph" <<endl;
	cout << "***********************************************************" <<endl ;
	for ( int i = 0 ; i < allNodes.size(); i++)
	{
		cout << allNodes[i].getid()  ;
		allNodes[i].show() ;
		cout << endl ;
	}
}

void graph::showForbidden()
{

	cout <<endl ;
	cout << "Adjacency list of nodes having some forbidden nodes in adjList" <<endl;
	cout << "***********************************************************" <<endl ;
	for ( int i = 0 ; i < allNodes.size(); i++)
	{
		if(allNodes[i].getAdjListSize() < 7)
		{
			cout << allNodes[i].getid()  ;
			allNodes[i].show() ;
			cout << endl ;
		}
	}
}

// In this procedure we assume that the graph already has the source and the destination and we build the graph.
// We also assume it already has the forbidden vector set up.

void graph :: buildAdjList()
{
	for ( int i = 0; i <10000; i++)
	{
		node tmpNode(i);

		tmpNode.createAdjList(forbiddenVector, forwardCost, backwardCost); // let us assume it creates everything.

		if(isIn(forbiddenVector, i)) tmpNode.setSafe(false);
		else 
		{
		tmpNode.setSafe(true);
		}
		addNode(tmpNode);
	}
}

void graph::addForbidden(int nid)
{
	forbiddenVector.push_back(nid);
}

bool graph::runBFS()
{
	queue<node> Q;
	for ( int k = 0; k<allNodes.size(); k++)
	{
		allNodes[k].setColor(0);
		allNodes[k].setDistance(INT_MAX);
		allNodes[k].setParent(-1);
	}
	node u;
	list<int>::iterator i;
	list<int> tmpList;

	allNodes[sourceNodeID].setDistance(0);
	allNodes[sourceNodeID].setParent(0);

	Q.push(allNodes[sourceNodeID]);
	while ( !Q.empty())
	{
		u = Q.front();
		Q.pop();

		tmpList = u.getAdjList();
		int dist = u.getDistance();
		int nid = u.getid();

		i = tmpList.begin();
		while ( i != tmpList.end())
		{

			if (allNodes[*i].getColor() == 0) 
			{
				allNodes[*i].setColor(1);
				allNodes[*i].setDistance(dist+1);
				allNodes[*i].setParent(nid);
				Q.push(allNodes[*i]);
			}
			i++;
		}
		allNodes[nid].setColor(2);
		if(nid == targetNodeID)
			return true;
	}
	return false;
}

bool graph::runDijkstra()
{

  for ( int k = 0; k<allNodes.size(); k++)
	{
		allNodes[k].setColor(0);
		allNodes[k].setDistance(INT_MAX);
		allNodes[k].setParent(-1);
		allNodes[k].setActive(true);
	}
 
  list<int>::iterator i;
  list<int>::iterator j;
  list<int> edgeList;
  list<int> costList;
  list<int> tmpList;

  allNodes[sourceNodeID].setDistance(0);
  allNodes[sourceNodeID].setParent(0);

	int minValue, minid, inactiveNodes, numRuns;
//	int tmpi, tmpj;
	minValue = 0 ;
	minid = 0;
	inactiveNodes = 0 ;
	numRuns = getSize()-getForbiddenNumber();

	while (inactiveNodes < numRuns)
	  {

		// extract the node with minimum distance value out of the queue
		minValue =  INT_MAX ;
		for ( int t = 0 ; t<allNodes.size(); t++)
		{ 
		if ((allNodes[t].getActive()== true ) &&   allNodes[t].getDistance() <minValue )
			{
			minValue = allNodes[t].getDistance() ;
			minid = t ;
			}
		}
		allNodes[minid].setActive(false) ;
		inactiveNodes++;
		edgeList = allNodes[minid].getAdjList();
		costList = allNodes[minid].getCostAdjList();

		if (!edgeList.empty())
		{

			i = edgeList.begin() ; 
			j = costList.begin();
		

		//	cout<<(*i)<<endl;
		//	cout<<(*j)<<endl;

			while(i != edgeList.end())
			{
			if ((minValue+(*j)) < allNodes[*i].getDistance())
				{
				allNodes[*i].setDistance(minValue+(*j)) ;
				allNodes[*i].setParent(minid) ;
				}
			i++ ;
			j++;
		//	tmpi = (*i);
		//	tmpj = (*j);
			}
		}
	}
	// Now we are done runnig Dijkstra's. Let us see the result.
	if(allNodes[targetNodeID].getParent() == -1)
		return false;
	else
		return true;
}


void graph::displayResults(ofstream & outputFile, int argumentCount)
{
		int nid = targetNodeID;
		int k,x1,x2,x3;
		vector<int> resultVector;
		int totalCost = allNodes[nid].getDistance();
		char * disp;
		int tmpid;
		bool notDone = true;
		while (notDone)
		{
			resultVector.push_back(nid);
			tmpid = allNodes[nid].getParent();
			nid = tmpid;
			if ( nid == sourceNodeID)
			{
				resultVector.push_back(nid);
				notDone = false;
			}
		}
		
		for ( int i = resultVector.size()-1; i >= 0; i--)
		{
			k = resultVector[i];
			x1 = k % 1000;
			x2 = x1 % 100;
			x3 = x2 % 10;
			if ( k >999) {
			  outputFile << k / 1000 << " " << x1 / 100 << " " << x2 / 10 << " " << x3 ;
			} else if ( k<=999 && k > 99) {
				outputFile << "0 " << x1 / 100 << " " << x2 / 10 << " " << x3;
			} else if ( k <= 99 && k > 9) {
				outputFile << "0 0 " << x2 / 10 << " " << x3;
			} else if ( k <= 9 && k>0) {
				outputFile << "0 0 0 " << x3;
			} else if ( k == 0) {
				outputFile << "0 0 0 0";
			}

			if ( i > 0)
			  outputFile << endl; //"--";
		}
		outputFile<<endl;
		outputFile <<"Minimum keypresses  required :"<<(resultVector.size()-1)<<endl;
		if ( argumentCount == 5)
		outputFile <<"Total cost : "<<totalCost<<endl;
		return;
}

void graph::displayParameters()
{
	cout<<"Source Node ID is :"<< sourceNodeID<<endl;
	cout<<"Target Node ID is :"<< targetNodeID<<endl;
	cout<<"There are "<< forbiddenVector.size()<<" forbidden configurations. These are :"<<endl;
	printForbidden();
}

void graph::printForbidden()
{
	for ( int i = 0 ; i< forbiddenVector.size(); i++)
		cout<< forbiddenVector[i]<< endl;
}

